/*
 * @(#)XmlConverter.java
 *
 * Copyright 2003, Tatsuya Asai, Hiroki Arimura, Takeaki Uno,
 * and Shin-ichi Nakano, all rights reserved.
 *
 * This software may be used freely for any purpose. However, when distributed,
 * the original source must be clearly stated, and, when the source code is
 * distributed, the copiright notice must be retained and any alterations in
 * the code must be clearly marked. No warranty is given regarding the quality
 * of this software.
 */

import javax.xml.parsers.*;
import org.xml.sax.*;
import org.xml.sax.helpers.*;
import java.util.*;
import java.io.*;
import unot.*;

/**
 * XmlConverterP.java
 * Created: Wed Sep  1 16:15:52 2004
 * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
 * @version $Revision: 1.6 $
 * A data converter that transforms an input file in XML format into
 * text file in parent tree form. 
 * The previous name is XmlConverterParentForm.
 */
public class XmlConverterP
{
	//Constants
	final static String USAGE_STRING =  
	  "Usage: java XmlConverterP [OPTIONS]  XMLFILE [XMLFILES ...]\n"+
	  "Options\n"+
	  "-d                   \t : indent each row by the node depth\n"+
	  "-w  | -word          \t : split a text into words (default)\n"+
	  "-nw | -noword        \t : donot split a text into words\n"+
	  "-o  | -out <outfile> \t : set the output to <outfile>. Otherwise, to stdout \n"+
	  "-m  | -multidoc <baseDepth> \t : treat an xml file as multiset of trees by \n"+
	  "considering each element of depth <baseDepth> as an indivisual document.\n"+
	  "";

	//Switches
	boolean ENABLE_OUTPUT_INDENT = false;
	boolean ENABLE_WORD_SPLIT    = false;
	static boolean ENABLE_MULTI_DOC  = false;
	static int baseElementDepth = 0; //default

	//
	int argc = 0;
	String outFile = null;

	//Parameter setting

	void setOutFile(String outFile) {
		this.outFile = outFile;
	};

	void setOutputIndentOn() {
		this.ENABLE_OUTPUT_INDENT = true;
	};

	void setWordSplitOn() {	this.ENABLE_WORD_SPLIT = true; };

	void setWordSplitOff() { this.ENABLE_WORD_SPLIT = false; };

	//Options

    void read_options(String[] args) {

		while (argc < args.length){

			//System.out.println("debug: option "+args[argc]);

			if (args[argc].charAt(0) == '-') { //An option starts
				if (args[argc].equals("-h") || args[argc].equals("-help")){
					argc++;
					System.out.println(USAGE_STRING);
					System.exit(0);
				}
				else if (args[argc].equals("-d") ||
						 args[argc].equals("-indent")){
					argc++;
					setOutputIndentOn(); 
				}
				else if (args[argc].equals("-w") ||
						 args[argc].equals("-word")){
					argc++;
					setWordSplitOn(); 
				}
				else if (args[argc].equals("-nw") ||
						 args[argc].equals("-noword")){
					argc++;
					setWordSplitOff(); 
				}
				else if (args[argc].equals("-o") ||
						 args[argc].equals("-out")){
					argc++;
					this.setOutFile(args[argc++]);
				}
				else if (args[argc].equals("-m") ||
						 args[argc].equals("-multidoc")){
					argc++;
					ENABLE_MULTI_DOC = true;
					baseElementDepth = Integer.parseInt(args[argc++]);
				}
				// Other case
				else {
					System.out.println("error: unknown option!: "+
									   args[argc]);
					System.exit(1);
				}
			}
			else {
				break;
			}

		}
	}


		/**
     * Describe <code>main</code> method here.
     *
     * @param args a <code>String[]</code> value
     */
    public static void main(String[] args)
    {
        if (args.length < 1) {
            System.out.println(USAGE_STRING);
            // $ java XmlConverterP XMLFILE | sort -n // to sort id :-)
            System.exit(0);
        }

        XmlConverterP conv = new XmlConverterP();

		conv.read_options(args);

        while (conv.argc < args.length) {
             conv.parseDoc(args[conv.argc++]);
        }
    }


    /**
     * Describe <code>parseDoc</code> method here.
     *
     * @param filename a <code>String</code> value
     */
    public void parseDoc(String filename)
    {
        SAXParserFactory factory = SAXParserFactory.newInstance();
		//DefaultHandler handler = new XmlParentFormHandler();
		XmlParentFormHandler handler = new XmlParentFormHandler();

        try {
            SAXParser parser = factory.newSAXParser();
            XMLReader reader = parser.getXMLReader();
            reader.setContentHandler(handler);
            reader.setErrorHandler(handler);

			//Setup FormHandler
			handler.ENABLE_OUTPUT_INDENT = this.ENABLE_OUTPUT_INDENT;
			
			if (this.outFile != null) 
				handler.setOutStream(new PrintStream(new FileOutputStream(outFile)));
			//Default is System.out
			if  (this.ENABLE_WORD_SPLIT)
				handler.setTextConverter(new WordSplitTextConverter(handler));
			else
				handler.setTextConverter(new SimpleTextConverter(handler));

			//Read input
            String path = new File(filename).getAbsolutePath();
            reader.parse("file:" + path);

        } catch (Exception e) {
            e.printStackTrace();
            System.exit(1);
        }
    }

} // XmlConverterP

/**
 * Describe class <code>UnotXmlHandler</code> here.
 *
 * @author <a href="mailto:s-fusa@i.kyushu-u.ac.jp">Shinji Fusanobu</a>
 * @version 1.0
 */
class XmlParentFormHandler extends DefaultHandler
{

    /** set true if you want to use serial node id number all over the documents. */
    public static boolean USE_SERIAL_NUMBER = true;
    public static final String VIRTUAL_GRAND_PARENT_ID = "-1";
	public static boolean ENABLE_OUTPUT_INDENT = false;
	public static boolean ENABLE_WORD_SPLIT    = false;

    static int freeNodeId = 0;
    StackList parentStack = null;
	StringBuffer textBuffer = new StringBuffer();
	PrintStream out = System.out;
	TextConverter textConverter = null; 
		
	int depth = 0;
	int numOfDecuments = 0;

	/** $B8=:_$NMWAG$N?<$5$rM?$($k!%%^%9%?!<:,$,(B -1$B$G!$:,$,(B0, $B$=$N;R$,(B+1$B$G$"$k!%(B */
	int getDepth() { return this.parentStack.size() - 1; }; 

    /**
     * Creates a new <code>XmlParentFormHandler</code> instance.
     *
     */
    public XmlParentFormHandler()
    {
        if (! USE_SERIAL_NUMBER) {
            this.freeNodeId = 0;
        }

		//this.parentStack = new ArrayList();
		this.parentStack = new StackList();
        this.parentStack.push(VIRTUAL_GRAND_PARENT_ID);
    }

	void setOutStream(PrintStream out) {
		this.out = out;
	};

	void setTextConverter(TextConverter textConverter) {
		this.textConverter = textConverter;
	};

    /**
     * Describe <code>startDocument</code> method here.
     *
     */
    public void startDocument()
    {
		this.depth = 0;
		this.out.println("@writer parent XmlConverterP.java"+"\n");
    }

    /**
     * Describe <code>startElement</code> method here.
     *
     * @param filename a <code>String</code> value
     * @param localName a <code>String</code> value
     * @param qualifiedName a <code>String</code> value
     * @param attributes an <code>Attributes</code> value
     */
    public void startElement(String filename, String localName,
                             String qualifiedName, Attributes attributes)
    {
		//discard text element
		if (textBuffer.length() > 0){
			textBuffer.setLength(0);
		}

		//Process the current element
		//writeStartAnAttrElem(qualifiedName);
		writeStartAnAttrElem(qualifiedName, "ELEM");

		for (int i = 0; i < attributes.getLength(); i++) {
			String name  = "@"+attributes.getQName(i);
			String value = attributes.getValue(i);

			writeStartAnAttrElem(name, "ATTR");
			writeStartAnAttrElem(value, "AVAL");
			writeEndAnAttrElem();
			writeEndAnAttrElem();
		}		
    }


    /**
     * Describe <code>endElement</code> method here.
     *
     * @param filename a <code>String</code> value
     * @param localName a <code>String</code> value
     * @param qualifiedName a <code>String</code> value
     */
    public void endElement(String filename, String localName, String qualifiedName)
    {
		//output text element
		if (textBuffer.length() > 0){
			//flushTextElem();
			this.textConverter.flush();
		}

		//end the current element
		writeEndAnAttrElem();
    }

	public void characters( char[] cbuf, int start, int len )
   {
	   //Accumulate characters
	   if (len > 0) {
		   textBuffer.append(cbuf, start, len );
	   }
   }

	/*
	 * Utilties
	 */

// 	private static Object getTopElement(List list) {
// 		return list.get(list.size() - 1);
// 	}
// 	private static Object popTopElement(List list) {
// 		Object obj = list.get(list.size() - 1);
// 		list.remove(list.size() - 1);
// 		return obj;
// 	}

	void writeStartAnAttrElem(String label){
		writeStartAnAttrElem(label, "");
	}

	void writeStartAnAttrElem(String label, String type){
		if (this.depth >= XmlConverterP.baseElementDepth) {

			//The specified root depth is encountered.
			if (this.depth == XmlConverterP.baseElementDepth) {
				//A new document
				this.out.println("@doc "+numOfDecuments);
				numOfDecuments++;

				//Initialize the name space for nodes
				this.parentStack = new StackList();
				this.parentStack.push(VIRTUAL_GRAND_PARENT_ID);
				this.freeNodeId = 0;
			}

			String selfId   = Integer.toString(this.freeNodeId);
			String parentId = (String) this.parentStack.getTop();
			//String parentId = (String) getTopElement(this.parentStack);

			if (ENABLE_OUTPUT_INDENT) {
				for (int i = 1; i < this.parentStack.size(); i++) 
					this.out.print("  ");
			}
			this.out.println(selfId+" "+parentId+" "+"\""+label+"\""+" "+type);
			this.parentStack.push(selfId);
		}

		this.depth++;
		//this.parentStack.add(selfId);
		this.freeNodeId++;
	}

	void writeEndAnAttrElem(){
		if (this.depth >= XmlConverterP.baseElementDepth) {
			//popTopElement(this.parentStack); 
			this.parentStack.pop();
		}
		this.depth--;
	}

	//Old version
	private void flushTextElem_NonSplit(){
		String text = textBuffer.toString();
		text.replace('\n', '#');
		text.replace('\r', '#');
		writeStartAnAttrElem(textBuffer.toString(), "TEXT");
		writeEndAnAttrElem();
		textBuffer.setLength(0);
	}	

	//private void flushTextElem_WordSplit(){
	private void flushTextElem(){
		StringReader _reader = new StringReader(textBuffer.toString());
		StreamTokenizer in = new StreamTokenizer(_reader);
		in.ordinaryChar('#');

		try {
			while (in.nextToken() != StreamTokenizer.TT_EOF) {
				if (in.ttype == StreamTokenizer.TT_WORD) {
					writeStartAnAttrElem(in.sval, "TEXT");
					writeEndAnAttrElem();

					//this.out.println("### "+in.sval);
				}
				else if (in.ttype == StreamTokenizer.TT_NUMBER) {
					writeStartAnAttrElem(String.valueOf((int)in.nval), "TEXT");
					writeEndAnAttrElem();
					//this.out.println("### "+in.sval);
				}
				else if (in.ttype == StreamTokenizer.TT_EOL) {
					//this.out.println("### EOL");
				}
			}
			//this.out.println("### EOF\n");
		}
		catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}

		textBuffer.setLength(0);
	}	

} // XmlParentFormHandler


abstract class TextConverter {
	XmlParentFormHandler handler = null;

	abstract void flush();
}

